#include "cleaning_robots_loop_functions.h"
#include <controllers/footbot_cleaning_robots/footbot_cleaning_robots.h>
#include <argos3/plugins/simulator/entities/cylinder_entity.h>
#include <argos3/core/simulator/simulator.h>
#include <argos3/core/utility/logging/argos_log.h>
#include <algorithm>
#include <cstring>
#include <cerrno>

/********************************************************************************/
/* Y-Coordinations of the working arena */
static const Real ARENA_MIN_Y  = -1.0f;
static const Real ARENA_MAX_Y  = 3.0f;
/********************************************************************************/
CCleaningRobotsLoopFunctions::CCleaningRobotsLoopFunctions() : m_rArenaX(0), m_rArenaY(0), m_bSetWorkedOnItem(false), 
							       m_pcFloor(NULL), m_RNG(NULL), m_LogCounter(1),
							       m_rPerformanceTimeUnit(0), m_rSingleRobotPerformance(0), NextCheck(10), NumofDeactivatedRobots(0), 
							       NumOfActivatedRobots(0), Init_Robots_Num(0){}
/********************************************************************************/

CCleaningRobotsLoopFunctions::~CCleaningRobotsLoopFunctions() {
  /* Nothing to do */
}
/********************************************************************************/

CCleaningRobotsLoopFunctions::SDirtItems::SDirtItems() :
  XDirtItemAmount(0), YDirtItemAmount(0), DirtItemRadius(0) {
}
/********************************************************************************/
CCleaningRobotsLoopFunctions::SRobotStatistic::SRobotStatistic() :
  RobotID(0), CleanedDirtItems(0) {
}

/********************************************************************************/

void CCleaningRobotsLoopFunctions::Init(TConfigurationNode& t_node) {
  TConfigurationNode& tCleaningRobots = GetNode(t_node, "cleaning_robots");
  /* Get a pointer to the floor entity */
  m_pcFloor = &GetSpace().GetFloorEntity();
  //random generator of positions
  m_RNG = CRandom::CreateRNG("argos");
  /* Get output file name from XML tree */
  GetNodeAttribute(tCleaningRobots, "single_dirt_radius", m_sDirtItem.DirtItemRadius);
  GetNodeAttribute(tCleaningRobots, "performance_time_unit", m_rPerformanceTimeUnit);
  GetNodeAttribute(tCleaningRobots, "arena_X", m_rArenaX);
  GetNodeAttribute(tCleaningRobots, "arena_Y", m_rArenaY);
  GetNodeAttribute(tCleaningRobots, "System_Performance_Log_File", m_fOutputSystemFileName);
  GetNodeAttribute(tCleaningRobots, "Active_Inactive_Robots_Log_File", m_fActiveInactiveRobotsFileName);
  GetNodeAttribute(tCleaningRobots, "init_num_robots", Init_Robots_Num);
  
	
  //amount of dirt items to be spread on the X axis and on the Y one
  m_sDirtItem.XDirtItemAmount = floor(m_rArenaX / m_sDirtItem.DirtItemRadius);
  m_sDirtItem.YDirtItemAmount = floor(m_rArenaY / m_sDirtItem.DirtItemRadius);
		
  double startXPoint = -2.0;
  double startYPoint = -1.0;
  double radius = m_sDirtItem.DirtItemRadius / 2;
  double firstCenterXPos = startXPoint + radius;
  double firstCenterYPos = startYPoint + radius;
  for (size_t i = 0; i < m_sDirtItem.XDirtItemAmount; ++i) {
    for (size_t j = 0; j < m_sDirtItem.YDirtItemAmount; ++j) {
      CVector2 newCenterPosition = CVector2(firstCenterXPos + (i * m_sDirtItem.DirtItemRadius),
					    firstCenterYPos + (j * m_sDirtItem.DirtItemRadius));
      m_sDirtItem.DirtItemsCenterPositions.push_back(newCenterPosition);
    }
  }
  //set the vector of the robots statistics
  CSpace::TMapPerType& cFBMap = GetSpace().GetEntitiesByType("foot-bot");
  for(CSpace::TMapPerType::iterator it = cFBMap.begin(); it != cFBMap.end(); ++it) {
    CFootBotEntity& footbotEntity = *any_cast<CFootBotEntity*>(it->second);
    CFootBotCleaningRobots& cController = dynamic_cast<CFootBotCleaningRobots&>(footbotEntity.GetControllableEntity().GetController());
			
    SRobotStatistic newRobotStatistics;
    newRobotStatistics.RobotID = atoi(cController.GetId().substr(3,
								 cController.GetId().length() - 3).c_str());
    newRobotStatistics.CleanedDirtItems = 0;
    m_sRobotsStatistics.push_back(newRobotStatistics);
			
  }
}

/********************************************************************************/

void CCleaningRobotsLoopFunctions::Reset() {
  //create random positions for this item equal to its amount
  m_sDirtItem.XDirtItemAmount = floor(m_rArenaX / m_sDirtItem.DirtItemRadius);
  m_sDirtItem.YDirtItemAmount = floor(m_rArenaY / m_sDirtItem.DirtItemRadius);
  double startXPoint = -2.0;
  double startYPoint = -1.0;
  Real radius = m_sDirtItem.DirtItemRadius / 2;
  for (UInt16 i = 0; i < m_sDirtItem.XDirtItemAmount; i++) {
    for (UInt16 j = 0; j < m_sDirtItem.YDirtItemAmount; j++) {
      CVector2 newCenterPosition = CVector2(startXPoint + radius + i * m_sDirtItem.DirtItemRadius, startYPoint
					    + radius + j * m_sDirtItem.DirtItemRadius);
      m_sDirtItem.DirtItemsCenterPositions.push_back(newCenterPosition);

    }
  }
}

/********************************************************************************/
void CCleaningRobotsLoopFunctions::Destroy() {
   
}
/********************************************************************************/
void CCleaningRobotsLoopFunctions::PreStep() {
  CSpace::TMapPerType& cFBMap = GetSpace().GetEntitiesByType("foot-bot");
  for(CSpace::TMapPerType::iterator it = cFBMap.begin(); it != cFBMap.end(); ++it) {
    CFootBotEntity& footbotEntity = *any_cast<CFootBotEntity*>(it->second);
    CVector2 cPos;
    cPos.Set(footbotEntity.GetEmbodiedEntity().GetOriginAnchor().Position.GetX(),
	     footbotEntity.GetEmbodiedEntity().GetOriginAnchor().Position.GetY());
    CFootBotCleaningRobots& cController = dynamic_cast<CFootBotCleaningRobots&>(footbotEntity.GetControllableEntity().GetController());
    CFootBotCleaningRobots::SRobotInfo& RobotInfo = cController.GetRobotInfo();
			
    /* If the robot is in the working arena and it is defected */
    if ((RobotInfo.sRobotState == CFootBotCleaningRobots::SRobotInfo::DEFECTED)&&(cPos.GetY()> -1.0)) {
       
    /* 1) Move the robot to the maintenance area to be repaired */ 
      bool bMoved = MoveEntity(
			       footbotEntity.GetEmbodiedEntity(),
			       CVector3(m_RNG->Uniform(CRange<Real>(-1.9,1.9)),  m_RNG->Uniform(CRange<Real>(-2.9,-1.1)), 0),
			       CQuaternion().FromEulerAngles(CRadians::ZERO,CRadians::ZERO,CRadians::ZERO), 
			       false 
			       );
	// If the move was not successful try again till it is      
		while(!bMoved)	  {
		bMoved = MoveEntity(
			       footbotEntity.GetEmbodiedEntity(),
			       CVector3(m_RNG->Uniform(CRange<Real>(-1.9,1.9)),  m_RNG->Uniform(CRange<Real>(-2.9,-1.1)), 0), 
			       CQuaternion().FromEulerAngles(CRadians::ZERO,CRadians::ZERO,CRadians::ZERO), 
			       false 
			       );
		
		}
      }	 else if (RobotInfo.sRobotState == CFootBotCleaningRobots::SRobotInfo::CLEANING) {
	 /* If the robot is in the cleaning state: 
	- compute the #squares cleaned per second
	- add it to the total #squares cleaned by this footbot 
	- determine the cleaned square */
	SRobotStatistic* RobotStatistic = GetRobotStatisticByRobotID(atoi(
									  cController.GetId().substr(3, cController.GetId().length()
												     - 3).c_str()));
	Real speeds[2]={0,0};
	speeds[0] = footbotEntity.GetWheeledEntity().GetWheelVelocity(0);
	speeds[1] = footbotEntity.GetWheeledEntity().GetWheelVelocity(1);
	Real NumberOfSquaresInSec = (float)(Abs(speeds[0])/m_sDirtItem.DirtItemRadius);
	RobotStatistic->CleanedDirtItems=RobotStatistic->CleanedDirtItems+NumberOfSquaresInSec;
	for (size_t pos_counter = 0; pos_counter < m_sDirtItem.DirtItemsCenterPositions.size(); pos_counter++) {
	  if ((cPos - m_sDirtItem.DirtItemsCenterPositions[pos_counter]).SquareLength()
	      < m_sDirtItem.DirtItemRadius * m_sDirtItem.DirtItemRadius) {	
	    CleanedDirtItemsCenterPositions.push_back(m_sDirtItem.DirtItemsCenterPositions[pos_counter]);
	   // m_pcFloor->SetChanged();
	  }

	}
	
      }
	/* If the robot is from the DEACTIVATED and the ACTIVATED */
	if(RobotInfo.add_me_to_deactivated_counter){
	NumofDeactivatedRobots = NumofDeactivatedRobots +1;
	RobotInfo.add_me_to_deactivated_counter = false;
	}
	if(RobotInfo.add_me_to_activated_counter){
	NumOfActivatedRobots = NumOfActivatedRobots + 1;
	RobotInfo.add_me_to_activated_counter=false;
	}
			
    }
    /* Log of the progressive number of the active and inactive robots */
    m_fActiveInactiveRobotsFile.open(m_fActiveInactiveRobotsFileName.c_str(), std::ios_base::app);
    if (m_fActiveInactiveRobotsFile.is_open()) {
		m_fActiveInactiveRobotsFile << GetSpace().GetSimulationClock() << ";" << NumofDeactivatedRobots <<";" << NumOfActivatedRobots << std::endl;
      }
    m_fActiveInactiveRobotsFile.close();
}
/********************************************************************************/

  void CCleaningRobotsLoopFunctions::PostStep() {
    /* Nothing to do */
  }

/********************************************************************************/

  CColor CCleaningRobotsLoopFunctions::GetFloorColor(const CVector2& c_position_on_plane) {
    /* Check if the given point is within the construction area */
    if(c_position_on_plane.GetY() < ARENA_MIN_Y){
      /* If we are on the maintenance area */
      return CColor::GRAY50;
    } else if (c_position_on_plane.GetY() > ARENA_MIN_Y){
      /* if we are on the working arena */
      for (UInt32 k = 0; k < CleanedDirtItemsCenterPositions.size(); k++) {
	if ((c_position_on_plane.GetX() >= CleanedDirtItemsCenterPositions[k].GetX() - m_sDirtItem.DirtItemRadius)
	    && ((c_position_on_plane.GetX() <= CleanedDirtItemsCenterPositions[k].GetX()
		 + m_sDirtItem.DirtItemRadius)) && ((c_position_on_plane.GetY()
						     >= CleanedDirtItemsCenterPositions[k].GetY() - m_sDirtItem.DirtItemRadius))
	    && ((c_position_on_plane.GetY() <= CleanedDirtItemsCenterPositions[k].GetY()
		 + m_sDirtItem.DirtItemRadius))) {
	  return CColor::WHITE;
	}
      }
    }
    return CColor::BLACK;
  }
/********************************************************************************/
  CCleaningRobotsLoopFunctions::SRobotStatistic* CCleaningRobotsLoopFunctions::GetRobotStatisticByRobotID(
													  UInt8 RobotID) {
    CCleaningRobotsLoopFunctions::SRobotStatistic* ItemStatisticToReturn = NULL;
    for (size_t i = 0; i < m_sRobotsStatistics.size(); i++) {
      if (m_sRobotsStatistics[i].RobotID == RobotID) {
	ItemStatisticToReturn = &m_sRobotsStatistics[i];
	break;
      }
    }
    return ItemStatisticToReturn;
  }
/********************************************************************************/
  bool CCleaningRobotsLoopFunctions::IsCenterPositionInList(CVector2 centerPostition,
							    std::vector<CVector2> positionList) {
    for (UInt32 posCounter = 0; posCounter < positionList.size(); posCounter++) {
      if (positionList[posCounter] == centerPostition) {
	return true;
      }
    }
    return false;
  }
  /********************************************************************************/
  /* This procedure is used to log the statistics of the experiment to the output files */
  void CCleaningRobotsLoopFunctions::LogStatistics(){
  CSpace::TMapPerType& cFBMap = GetSpace().GetEntitiesByType("foot-bot");
        /* The total area to clean is the dirty area multiplied by the number of robots,
	 /as it is possible for the robot to clean a cleaned square */
      UInt32 TotolSquaresAvailableToClean = m_sDirtItem.DirtItemsCenterPositions.size();
      UInt32 TotalPerformance = 0;
      /* Sum up the performance achieved by each robot to obtain the total performance */
      for(CSpace::TMapPerType::iterator it = cFBMap.begin(); it != cFBMap.end(); ++it) {
	CFootBotEntity& footbotEntity = *any_cast<CFootBotEntity*>(it->second);
	CFootBotCleaningRobots& cController = dynamic_cast<CFootBotCleaningRobots&>(footbotEntity.GetControllableEntity().GetController());
	SRobotStatistic* RobotStatistic = GetRobotStatisticByRobotID(
								     atoi(
									  cController.GetId().substr(3, cController.GetId().length()
												     - 3).c_str()));
	TotalPerformance = TotalPerformance + floor(RobotStatistic->CleanedDirtItems);
      }
      /* save system performance statistics to the output file */
      m_fOutputSystemFile.open(m_fOutputSystemFileName.c_str(), std::ios_base::app);
      if (m_fOutputSystemFile.is_open()) {
	/* 0.0009 = 0.03 * 0.03 (the area of one square)*/
	m_fOutputSystemFile << m_sRobotsStatistics.size() << ";"
			    << TotalPerformance * 0.0009<< std::endl;
      }
      m_fOutputSystemFile.close();
      
  }
/********************************************************************************/
  bool CCleaningRobotsLoopFunctions::IsExperimentFinished() {

	/* Stop the experiment when all robots become inactive (extinction time) */
	if(Init_Robots_Num + NumOfActivatedRobots - NumofDeactivatedRobots<=0){
	LogStatistics();
	return true;
	}else{
	return false;
	}

  }
/********************************************************************************/
  /* Register this loop functions into the ARGoS plugin system */
  REGISTER_LOOP_FUNCTIONS(CCleaningRobotsLoopFunctions, "cleaning_robots_loop_functions");
