#include <argos3/core/simulator/loop_functions.h>
#include <argos3/plugins/robots/foot-bot/simulator/footbot_entity.h>
#include <argos3/core/simulator/entity/floor_entity.h>
#include <argos3/core/utility/math/rng.h>
#include <fstream>

using namespace argos;

class CCleaningRobotsLoopFunctions : public CLoopFunctions {
public:

	struct  SDirtItems{
		UInt16 XDirtItemAmount;
		UInt16 YDirtItemAmount;
		double DirtItemRadius;
		std::vector<CVector2> DirtItemsCenterPositions;
		SDirtItems();
	};
	struct SRobotStatistic {
		UInt8 RobotID;
		std::vector<CVector2> DirtItemsCenterPositions;
		Real CleanedDirtItems;
		SRobotStatistic();
	};

public:

   /**
    * Class constructor
    */
   CCleaningRobotsLoopFunctions();

   /** 
    * Class destructor
    */
   virtual ~CCleaningRobotsLoopFunctions();

   /**
    * Initializes the experiment.
    * It is executed once at the beginning of the experiment, i.e., when ARGoS is launched.
    * @param t_tree The parsed XML tree corresponding to the <loop_functions> section.
    */
   virtual void Init(TConfigurationNode& t_tree);

   /**
    * Resets the experiment to the state it was right after Init() was called.
    * It is executed every time you press the 'reset' button in the GUI.
    */
   virtual void Reset();

   /**
    * Undoes whatever Init() did.
    * It is executed once when ARGoS has finished the experiment.
    */
   virtual void Destroy();

   /**
    * Performs actions right before a simulation step is executed.
    */
   virtual void PreStep();

   /**
    * Performs actions right after a simulation step is executed.
    */
   virtual void PostStep();

   /**
    * Returns the color of the floor at the specified point on.
    * @param c_position_on_plane The position at which you want to get the color.
    * @see CColor
    */
   virtual CColor GetFloorColor(const CVector2& c_position_on_plane);

	/**
    * Get the robots statistic information by robot ID
    */
  virtual SRobotStatistic* GetRobotStatisticByRobotID(UInt8 RobotID);
  virtual void LogStatistics();
	virtual bool IsCenterPositionInList(CVector2 centerPosition,std::vector<CVector2> positionsList);
	virtual bool IsExperimentFinished();

private:


   /**
    * The path of the output file.
    */
   std::string m_strOutFile;

   /**
    * The stream associated to the output file.
    */
   std::ofstream m_cOutFile;

   /**
    * This vector contains a list of positions of objects in the construction area
    */
  // std::vector<CVector3> m_vecConstructionObjectsInArea;
  /*The X and Y projection of the arena*/
	Real m_rArenaX, m_rArenaY;
	 /*Dirt Item*/
	SDirtItems m_sDirtItem;
  /* A flag to indicate that it has been worked on this item */
	bool m_bSetWorkedOnItem;
 /* Floor entity */
	CFloorEntity* m_pcFloor;
	CRandom::CRNG* m_RNG;
	std::vector<SRobotStatistic> m_sRobotsStatistics;
	size_t m_LogCounter;
	std::string m_sDistributionOfObjects;
	Real m_rPerformanceTimeUnit;
	Real m_rSingleRobotPerformance;
	std::vector<CVector2> CleanedDirtItemsCenterPositions;
	std::ofstream m_fOutputSystemFile;
	std::string m_fOutputSystemFileName;
	std::ofstream m_fActiveInactiveRobotsFile;
	std::string m_fActiveInactiveRobotsFileName;
	Real  NextCheck;
	Real NumofDeactivatedRobots;
	Real NumOfActivatedRobots;
	Real Init_Robots_Num;
};
